# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import unittest
from io import StringIO
from types import SimpleNamespace

import lsst.pex.config.configurableActions.tests as actionTests
from lsst.pex.config import FieldValidationError
from lsst.pex.config.configurableActions.tests import ActionTest1, ActionTest2, ActionTest3


class ConfigurableActionsTestCase(unittest.TestCase):
    """Test fo ConfigurableActions."""

    def _createConfig(self, default=None, singleDefault=None):
        class NewTestConfig(actionTests.TestConfig):
            def setDefaults(self):
                super().setDefaults()
                if default is not None:
                    for k, v in default.items():
                        setattr(self.actions, k, v)
                if singleDefault is not None:
                    self.singleAction = singleDefault

        return NewTestConfig

    def testConfigInstantiation(self):
        # This will raise if there is an issue instantiating something
        configClass = self._createConfig()
        config = configClass()
        self.assertTrue(hasattr(config, "actions"))
        self.assertTrue(hasattr(config, "singleAction"))

        # test again with default values
        configClass = self._createConfig(default={"test1": ActionTest1}, singleDefault=ActionTest1)
        config = configClass()

        # verify the defaults were created
        self.assertTrue(hasattr(config.actions, "test1"))
        self.assertTrue(hasattr(config.actions.test1, "var"))
        self.assertEqual(config.actions.test1.var, 0)

        self.assertTrue(hasattr(config.singleAction, "var"))
        self.assertEqual(config.singleAction.var, 0)

    def testAssignment(self):
        # Struct actions
        # Test that a new action can be added with assignment
        configClass = self._createConfig(default={"test1": ActionTest1})
        config = configClass()
        config.actions.test2 = ActionTest2

        self.assertEqual(tuple(config.actions.fieldNames), ("test1", "test2"))
        self.assertEqual(config.actions.test2.var, 1)

        # verify the same as above, but assigning with instances
        configClass = self._createConfig(default={"test1": ActionTest1})
        config = configClass()
        config.actions.test3 = ActionTest3()

        self.assertEqual(tuple(config.actions.fieldNames), ("test1", "test3"))
        self.assertEqual(config.actions.test3.var, 3)

        # The following is designed to support pipeline config setting
        # Test assignment using the update accessor
        configClass = self._createConfig(default={"test1": ActionTest1})
        config = configClass()
        config.actions.update = {"test2": ActionTest2, "test3": ActionTest3}

        self.assertEqual(tuple(config.actions.fieldNames), ("test1", "test2", "test3"))

        configClass = self._createConfig(default={"test1": ActionTest1})
        configClass2 = self._createConfig(default={"test2": ActionTest2, "test3": ActionTest3})
        config = configClass()
        config2 = configClass2()
        config.actions.update = config2.actions

        self.assertEqual(tuple(config.actions.fieldNames), ("test1", "test2", "test3"))

        # verify tha the update interface cannot be used to assign invalid
        # identifiers
        configClass = self._createConfig()
        config = configClass()
        with self.assertRaises(ValueError):
            config.actions.update = {"name with space": ActionTest2}

        with self.assertRaises(ValueError):
            config.actions.update = {"9leading_number": ActionTest2}

        # Test remove "assignment" using the remove accessor
        configClass = self._createConfig(
            default={"test1": ActionTest1, "test2": ActionTest2, "test3": ActionTest3}
        )
        config = configClass()
        config.actions.remove = ("test1", "test2")
        self.assertEqual(tuple(config.actions.fieldNames), ("test3",))

        configClass = self._createConfig(
            default={"test1": ActionTest1, "test2": ActionTest2, "test3": ActionTest3}
        )
        config = configClass()
        config.actions.remove = "test1"
        self.assertEqual(tuple(config.actions.fieldNames), ("test2", "test3"))

        # singleAction
        # Test that an action can be reassigned
        configClass = self._createConfig(singleDefault=ActionTest1)
        config = configClass()
        self.assertEqual(config.singleAction(), 0)

        config.singleAction = ActionTest2
        self.assertEqual(config.singleAction(), 1)

        config.singleAction = ActionTest3()
        self.assertEqual(config.singleAction(), 3)

        # Verify that ConfigurableActionStructField can be assigned to with
        # a ConfigurableActionStruct, SimpleNamespace
        otherConfigClass = self._createConfig(
            default={"test1": ActionTest1(var=1), "test2": ActionTest2(var=2)}
        )
        assignSource1 = otherConfigClass().actions
        assignSource2 = SimpleNamespace(test1=ActionTest1(var=1), test2=ActionTest2(var=2))

        for source in (assignSource1, assignSource2):
            configClass = self._createConfig()
            config = configClass()
            config.actions = source

            self.assertEqual(tuple(config.actions.fieldNames), ("test1", "test2"))
            self.assertEqual((config.actions.test1.var, config.actions.test2.var), (1, 2))

        # Fail if assigment is ConfigurableActionStructField
        with self.assertRaises(ValueError):
            configClass = self._createConfig()
            config = configClass()
            config.actions = otherConfigClass.actions

        # Fail if assignment is some other type
        with self.assertRaises(ValueError):
            configClass = self._createConfig()
            config = configClass()
            config.actions = {}

    def testValidate(self):
        configClass = self._createConfig(
            default={"test1": ActionTest1, "test2": ActionTest2, "test3": ActionTest3},
            singleDefault=ActionTest1,
        )
        config = configClass()
        config.validate()

    def testFreeze(self):
        configClass = self._createConfig(
            default={"test1": ActionTest1, "test2": ActionTest2}, singleDefault=ActionTest1
        )
        config = configClass()
        config.freeze()

        with self.assertRaises(FieldValidationError):
            config.actions.test3 = ActionTest3

        with self.assertRaises(FieldValidationError):
            config.actions.test1.var = 2

        with self.assertRaises(FieldValidationError):
            config.actions.test2.var = 0

        with self.assertRaises(FieldValidationError):
            config.singleAction = ActionTest2

        with self.assertRaises(FieldValidationError):
            config.singleAction.var = 3

    def testCompare(self):
        configClass = self._createConfig(
            default={"test1": ActionTest1, "test2": ActionTest2}, singleDefault=ActionTest1
        )
        config = configClass()
        config2 = configClass()

        self.assertTrue(config.compare(config2))

        # Test equality fail for ConfigurableActionsStructField
        config3 = configClass()
        config3.actions.test1.var = 99
        self.assertFalse(config.compare(config3))

        # Test equality fail for ConfigurableActionsField
        config4 = configClass()
        config4.singleAction.var = 99
        self.assertFalse(config.compare(config4))

    def testSave(self):
        # This method will also test rename, as it is part of the
        # implementation in pex_config
        ioObject = StringIO()
        config = actionTests.TestConfig()
        config.actions.test1 = ActionTest1
        config.actions.test2 = ActionTest2
        config.singleAction = actionTests.TestDivideAction(
            colA=actionTests.TestSingleColumnAction(column="a"),
            colB=actionTests.TestSingleColumnAction(column="b"),
        )

        config.saveToStream(ioObject)
        string1 = ioObject.getvalue()
        loadedConfig = actionTests.TestConfig()
        loadedConfig.loadFromStream(string1)
        self.assertTrue(config.compare(loadedConfig), msg=f"{config} != {loadedConfig}")
        # Be sure that the fields are actually there
        self.assertEqual(loadedConfig.actions.test1.var, 0)
        self.assertEqual(loadedConfig.singleAction.colA.column, "a")
        self.assertEqual(loadedConfig.singleAction.colB.column, "b")
        # Save an equivalent struct with fields originally ordered differently,
        # check that the saved form is the same (via deterministic sorting).
        config2 = actionTests.TestConfig()
        config2.actions.test2 = ActionTest2
        config2.actions.test1 = ActionTest1
        config2.singleAction = actionTests.TestDivideAction(
            colB=actionTests.TestSingleColumnAction(column="b"),
            colA=actionTests.TestSingleColumnAction(column="a"),
        )
        ioObject2 = StringIO()
        config2.saveToStream(ioObject2)
        self.maxDiff = None
        self.assertEqual(string1, ioObject2.getvalue())

    def testToDict(self):
        """Test the toDict interface."""
        configClass = self._createConfig(default={"test1": ActionTest1}, singleDefault=ActionTest1)
        config = configClass()
        self.assertEqual(config.toDict(), {"actions": {"test1": {"var": 0}}, "singleAction": {"var": 0}})

    def test_copy(self):
        """Test copying a Config with ConfigurableActions."""
        config_cls = self._createConfig(default={"test1": ActionTest1}, singleDefault=ActionTest1)
        config = config_cls()
        copy1 = config.copy()
        copy1.actions.test2 = ActionTest2
        self.assertEqual(tuple(copy1.actions.fieldNames), ("test1", "test2"))
        self.assertEqual(copy1.actions.test2.var, 1)
        self.assertEqual(tuple(config.actions.fieldNames), ("test1",))
        copy1.actions.test1.var = 50
        self.assertEqual(copy1.actions.test1.var, 50)
        self.assertEqual(config.actions.test1.var, 0)
        copy2 = copy1.copy()
        copy2.actions.test3 = ActionTest3()
        self.assertEqual(tuple(copy2.actions.fieldNames), ("test1", "test2", "test3"))
        self.assertEqual(tuple(copy1.actions.fieldNames), ("test1", "test2"))
        self.assertEqual(tuple(config.actions.fieldNames), ("test1",))
        self.assertEqual(copy2.actions.test3.var, 3)


if __name__ == "__main__":
    unittest.main()



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["StackFrame", "getCallStack", "getCallerFrame", "getStackFrame"]

import inspect
import linecache


def getCallerFrame(relative=0):
    """Get the frame for the user's caller.

    Parameters
    ----------
    relative : `int`, optional
        Number of frames (0 or more) above the caller to retrieve.
        Default is 0.

    Returns
    -------
    frame : `types.FrameType`
        Frame for the caller.

    Notes
    -----
    This function is excluded from the frame.
    """
    frame = inspect.currentframe().f_back.f_back  # Our caller's caller
    for _ in range(relative):
        frame = frame.f_back
    return frame


def getStackFrame(relative=0):
    """Get the `StackFrame` for the user's caller.

    Parameters
    ----------
    relative : `int`, optional
        Number of frames (0 or more) above the caller to retrieve.

    Returns
    -------
    frame : `StackFrame`
        Stack frame for the caller.
    """
    frame = getCallerFrame(relative + 1)
    return StackFrame.fromFrame(frame)


class StackFrame:
    """A single element of the stack trace.

    This differs slightly from the standard system mechanisms for getting a
    stack trace by the fact that it does not look up the source code until it
    is absolutely necessary, reducing the I/O.

    Parameters
    ----------
    filename : `str`
        Name of file containing the code being executed.
    lineno : `int`
        Line number of file being executed.
    function : `str`
        Function name being executed.
    content : `str`, optional
        The actual content being executed. If not provided, it will be loaded
        from the file.

    Notes
    -----
    This differs slightly from the standard system mechanisms for getting a
    stack trace by the fact that it does not look up the source code until it
    is absolutely necessary, reducing the I/O.

    See Also
    --------
    getStackFrame
    """

    _STRIP = "/python/lsst/"
    """String to strip from the ``filename`` in the constructor."""

    def __init__(self, filename, lineno, function, content=None):
        loc = filename.rfind(self._STRIP)
        if loc > 0:
            filename = filename[loc + len(self._STRIP) :]
        self.filename = filename
        self.lineno = lineno
        self.function = function
        self._content = content

    @property
    def content(self):
        """Content being executed (loaded on demand) (`str`)."""
        if self._content is None:
            self._content = linecache.getline(self.filename, self.lineno).strip()
        return self._content

    @classmethod
    def fromFrame(cls, frame):
        """Construct from a Frame object.

        Parameters
        ----------
        frame : `types.FrameType`
            Frame object to interpret, such as from `inspect.currentframe`.

        Returns
        -------
        stackFrame : `StackFrame`
            A `StackFrame` instance.

        Examples
        --------
        `inspect.currentframe` provides a Frame object. This is a convenience
        constructor to interpret that Frame object:

        >>> import inspect
        >>> stackFrame = StackFrame.fromFrame(inspect.currentframe())
        """
        filename = frame.f_code.co_filename
        lineno = frame.f_lineno
        function = frame.f_code.co_name
        return cls(filename, lineno, function)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.filename}, {self.lineno}, {self.function})"

    def format(self, full=False):
        """Format for printing.

        Parameters
        ----------
        full : `bool`, optional
            If `True`, output includes the conentent (`StackFrame.content`)
            being executed. Default is `False`.

        Returns
        -------
        result : `str`
            Formatted string.
        """
        result = f"  File {self.filename}:{self.lineno} ({self.function})"
        if full:
            result += f"\n    {self.content}"
        return result


def getCallStack(skip=0):
    """Retrieve the call stack for the caller.

    Parameters
    ----------
    skip : `int`, non-negative
        Number of stack frames above caller to skip.

    Returns
    -------
    output : `list` of `StackFrame`
        The call stack. The `list` is ordered with the most recent frame to
        last.

    Notes
    -----
    This function is excluded from the call stack.
    """
    frame = getCallerFrame(skip + 1)
    stack = []
    while frame:
        stack.append(StackFrame.fromFrame(frame))
        frame = frame.f_back
    return list(reversed(stack))



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ("makePropertySet",)

try:
    import lsst.daf.base as dafBase
except ImportError:
    dafBase = None


def _helper(ps, prefix, dict_):
    for k, v in dict_.items():
        name = prefix + "." + k if prefix is not None else k
        if isinstance(v, dict):
            _helper(ps, name, v)
        elif v is not None:
            ps.set(name, v)


def makePropertySet(config):
    """Convert a configuration into a `lsst.daf.base.PropertySet`.

    Parameters
    ----------
    config : `lsst.pex.config.Config`
        Configuration instance.

    Returns
    -------
    propertySet : `lsst.daf.base.PropertySet`
        A `~lsst.daf.base.PropertySet` that is equivalent to the ``config``
        instance. If ``config`` is `None` then this return value is also
        `None`.

    See Also
    --------
    lsst.daf.base.PropertySet
    """
    if dafBase is None:
        raise RuntimeError("lsst.daf.base is not available")

    if config is not None:
        ps = dafBase.PropertySet()
        _helper(ps, None, config.toDict())
        return ps
    else:
        return None



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ("Registry", "RegistryField", "makeRegistry", "registerConfig", "registerConfigurable")

import collections.abc
import copy

from .config import Config, FieldValidationError, _typeStr
from .configChoiceField import ConfigChoiceField, ConfigInstanceDict


class ConfigurableWrapper:
    """A wrapper for configurables.

    Used for configurables that don't contain a ``ConfigClass`` attribute,
    or contain one that is being overridden.

    Parameters
    ----------
    target : configurable class
        Target class.
    ConfigClass : `type`
        Config class.
    """

    def __init__(self, target, ConfigClass):
        self.ConfigClass = ConfigClass
        self._target = target

    def __call__(self, *args, **kwargs):
        return self._target(*args, **kwargs)


class Registry(collections.abc.Mapping):
    """A base class for global registries, which map names to configurables.

    A registry acts like a read-only dictionary with an additional `register`
    method to add targets. Targets in the registry are configurables (see
    *Notes*).

    Parameters
    ----------
    configBaseType : `lsst.pex.config.Config`-type
        The base class for config classes in the registry.

    Notes
    -----
    A configurable is a callable with call signature ``(config, *args)``
    Configurables typically create an algorithm or are themselves the
    algorithm. Often configurables are `lsst.pipe.base.Task` subclasses, but
    this is not required.

    A ``Registry`` has these requirements:

    - All configurables added to a particular registry have the same call
      signature.
    - All configurables in a registry typically share something important
      in common. For example, all configurables in ``psfMatchingRegistry``
      return a PSF matching class that has a ``psfMatch`` method with a
      particular call signature.

    Examples
    --------
    This examples creates a configurable class ``Foo`` and adds it to a
    registry. First, creating the configurable:

    >>> from lsst.pex.config import Registry, Config
    >>> class FooConfig(Config):
    ...     val = Field(dtype=int, default=3, doc="parameter for Foo")
    >>> class Foo:
    ...     ConfigClass = FooConfig
    ...
    ...     def __init__(self, config):
    ...         self.config = config
    ...
    ...     def addVal(self, num):
    ...         return self.config.val + num

    Next, create a ``Registry`` instance called ``registry`` and register the
    ``Foo`` configurable under the ``"foo"`` key:

    >>> registry = Registry()
    >>> registry.register("foo", Foo)
    >>> print(list(registry.keys()))
    ["foo"]

    Now ``Foo`` is conveniently accessible from the registry itself.

    Finally, use the registry to get the configurable class and create an
    instance of it:

    >>> FooConfigurable = registry["foo"]
    >>> foo = FooConfigurable(FooConfigurable.ConfigClass())
    >>> foo.addVal(5)
    8
    """

    def __init__(self, configBaseType=Config):
        if not issubclass(configBaseType, Config):
            raise TypeError(f"configBaseType={_typeStr(configBaseType)} must be a subclass of Config")
        self._configBaseType = configBaseType
        self._dict = {}

    def register(self, name, target, ConfigClass=None):
        """Add a new configurable target to the registry.

        Parameters
        ----------
        name : `str`
            Name that the ``target`` is registered under. The target can
            be accessed later with `dict`-like patterns using ``name`` as
            the key.
        target : obj
            A configurable type, usually a subclass of `lsst.pipe.base.Task`.
        ConfigClass : `lsst.pex.config.Config`-type, optional
            A subclass of `lsst.pex.config.Config` used to configure the
            configurable. If `None` then the configurable's ``ConfigClass``
            attribute is used.

        Raises
        ------
        RuntimeError
            Raised if an item with ``name`` is already in the registry.
        AttributeError
            Raised if ``ConfigClass`` is `None` and ``target`` does not have
            a ``ConfigClass`` attribute.

        Notes
        -----
        If ``ConfigClass`` is provided then the ``target`` configurable is
        wrapped in a new object that forwards function calls to it. Otherwise
        the original ``target`` is stored.
        """
        if name in self._dict:
            raise RuntimeError(f"An item with name {name!r} already exists")
        if ConfigClass is None:
            wrapper = target
        else:
            wrapper = ConfigurableWrapper(target, ConfigClass)
        if not issubclass(wrapper.ConfigClass, self._configBaseType):
            raise TypeError(
                f"ConfigClass={_typeStr(wrapper.ConfigClass)} is not a subclass of "
                f"{_typeStr(self._configBaseType)!r}"
            )
        self._dict[name] = wrapper

    def __getitem__(self, key):
        return self._dict[key]

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def __contains__(self, key):
        return key in self._dict

    def makeField(self, doc, default=None, optional=False, multi=False, on_none=None):
        """Create a `RegistryField` configuration field from this registry.

        Parameters
        ----------
        doc : `str`
            A description of the field.
        default : object, optional
            The default target for the field.
        optional : `bool`, optional
            When `False`, `lsst.pex.config.Config.validate` fails if the
            field's value is `None`.
        multi : `bool`, optional
            A flag to allow multiple selections in the `RegistryField` if
            `True`.
        on_none : `Callable`, optional
            A callable that should be invoked when ``apply`` is called but the
            selected name or names is `None`.  Will be passed the field
            attribute proxy (`RegistryInstanceDict`) and then all positional
            and keyword arguments passed to ``apply``.

        Returns
        -------
        field : `lsst.pex.config.RegistryField`
            `~lsst.pex.config.RegistryField` Configuration field.
        """
        return RegistryField(doc, self, default, optional, multi, on_none=on_none)


class RegistryAdaptor(collections.abc.Mapping):
    """Private class that makes a `Registry` behave like the thing a
    `~lsst.pex.config.ConfigChoiceField` expects.

    Parameters
    ----------
    registry : `Registry`
        `Registry` instance.
    """

    def __init__(self, registry):
        self.registry = registry

    def __getitem__(self, k):
        return self.registry[k].ConfigClass

    def __iter__(self):
        return iter(self.registry)

    def __len__(self):
        return len(self.registry)

    def __contains__(self, k):
        return k in self.registry


class RegistryInstanceDict(ConfigInstanceDict):
    """Dictionary of instantiated configs, used to populate a `RegistryField`.

    Parameters
    ----------
    config : `lsst.pex.config.Config`
        Configuration instance.
    field : `RegistryField`
        Configuration field.
    """

    def __init__(self, config, field):
        ConfigInstanceDict.__init__(self, config, field)
        self.registry = field.registry

    def _getTarget(self):
        if self._field.multi:
            raise FieldValidationError(
                self._field, self._config, "Multi-selection field has no attribute 'target'"
            )
        return self.types.registry[self._selection]

    target = property(_getTarget)

    def _getTargets(self):
        if not self._field.multi:
            raise FieldValidationError(
                self._field, self._config, "Single-selection field has no attribute 'targets'"
            )
        return [self.types.registry[c] for c in self._selection]

    targets = property(_getTargets)

    def apply(self, *args, **kwargs):
        """Call the active target(s) with the active config as a keyword arg.

        Parameters
        ----------
        *args, **kwargs : `~typing.Any
            Additional arguments will be passed on to the configurable
            target(s).

        Returns
        -------
        result
            If this is a single-selection field, the return value from calling
            the target. If this is a multi-selection field, a list thereof.
        """
        if self.active is None:
            if self._field._on_none is not None:
                return self._field._on_none(self, *args, **kwargs)
            msg = "No selection has been made.  Options: {}".format(" ".join(self.types.registry.keys()))
            raise FieldValidationError(self._field, self._config, msg)
        return self.apply_with(self._selection, *args, **kwargs)

    def apply_with(self, selection, *args, **kwargs):
        """Call named target(s) with the corresponding config as a keyword
        arg.

        Parameters
        ----------
        selection : `str` or `~collections.abc.Iterable` [ `str` ]
            Name or names of targets, depending on whether ``multi=True``.
        *args, **kwargs
            Additional arguments will be passed on to the configurable
            target(s).

        Returns
        -------
        result
            If this is a single-selection field, the return value from calling
            the target. If this is a multi-selection field, a list thereof.

        Notes
        -----
        This method ignores the current selection in the ``name`` or ``names``
        attribute, which is usually not what you want.  This method is most
        useful in ``on_none`` callbacks provided at field construction, which
        allow a context-dependent default to be used when no selection is
        configured.
        """
        if self._field.multi:
            retvals = []
            for c in selection:
                retvals.append(self.types.registry[c](*args, config=self[c], **kwargs))
            return retvals
        else:
            return self.types.registry[selection](*args, config=self[selection], **kwargs)

    def __setattr__(self, attr, value):
        if attr == "registry":
            object.__setattr__(self, attr, value)
        else:
            ConfigInstanceDict.__setattr__(self, attr, value)


class RegistryField(ConfigChoiceField):
    """A configuration field whose options are defined in a `Registry`.

    Parameters
    ----------
    doc : `str`
        A description of the field.
    registry : `Registry`
        The registry that contains this field.
    default : `str`, optional
        The default target key.
    optional : `bool`, optional
        When `False`, `lsst.pex.config.Config.validate` fails if the field's
        value is `None`.
    multi : `bool`, optional
        If `True`, the field allows multiple selections. The default is
        `False`.
    on_none : `Callable`, optional
        A callable that should be invoked when ``apply`` is called but the
        selected name or names is `None`.  Will be passed the field attribute
        proxy (`RegistryInstanceDict`) and then all positional and keyword
        arguments passed to ``apply``.

    See Also
    --------
    ChoiceField
    ConfigChoiceField
    ConfigDictField
    ConfigField
    ConfigurableField
    DictField
    Field
    ListField
    RangeField
    """

    instanceDictClass = RegistryInstanceDict
    """Class used to hold configurable instances in the field.
    """

    def __init__(self, doc, registry, default=None, optional=False, multi=False, on_none=None):
        types = RegistryAdaptor(registry)
        self.registry = registry
        self._on_none = on_none
        ConfigChoiceField.__init__(self, doc, types, default, optional, multi)

    def __deepcopy__(self, memo):
        """Customize deep-copying, want a reference to the original registry.

        WARNING: this must be overridden by subclasses if they change the
        constructor signature!
        """
        other = type(self)(
            doc=self.doc,
            registry=self.registry,
            default=copy.deepcopy(self.default),
            optional=self.optional,
            multi=self.multi,
            on_none=self._on_none,
        )
        other.source = self.source
        return other


def makeRegistry(doc, configBaseType=Config):
    """Create a `Registry`.

    Parameters
    ----------
    doc : `str`
        Docstring for the created `Registry` (this is set as the ``__doc__``
        attribute of the `Registry` instance.
    configBaseType : `lsst.pex.config.Config`-type
        Base type of config classes in the `Registry`.

    Returns
    -------
    registry : `Registry`
        Registry with ``__doc__`` and ``_configBaseType`` attributes
        set.
    """
    cls = type("Registry", (Registry,), {"__doc__": doc})
    return cls(configBaseType=configBaseType)


def registerConfigurable(name, registry, ConfigClass=None):
    """Add a class as a configurable in a `Registry` instance.

    Parameters
    ----------
    name : `str`
        Name of the target (the decorated class) in the ``registry``.
    registry : `Registry`
        The `Registry` instance that the decorated class is added to.
    ConfigClass : `lsst.pex.config.Config`-type, optional
        Config class associated with the configurable. If `None`, the class's
        ``ConfigClass`` attribute is used instead.

    See Also
    --------
    registerConfig

    Notes
    -----
    Internally, this decorator runs `Registry.register`.
    """

    def decorate(cls):
        registry.register(name, target=cls, ConfigClass=ConfigClass)
        return cls

    return decorate


def registerConfig(name, registry, target):
    """Add a class as a ``ConfigClass`` in a `Registry` and
    associate it with the given configurable.

    Parameters
    ----------
    name : `str`
        Name of the ``target`` in the ``registry``.
    registry : `Registry`
        The registry containing the ``target``.
    target : obj
        A configurable type, such as a subclass of `lsst.pipe.base.Task`.

    See Also
    --------
    registerConfigurable

    Notes
    -----
    Internally, this decorator runs `Registry.register`.
    """

    def decorate(cls):
        registry.register(name, target=target, ConfigClass=cls)
        return cls

    return decorate



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ["DictField"]

import collections.abc
import weakref
from collections.abc import Iterator, Mapping
from typing import Any, ForwardRef, Generic, TypeVar, cast

from .callStack import StackFrame, getCallStack, getStackFrame
from .comparison import compareScalars, getComparisonName
from .config import (
    Config,
    Field,
    FieldValidationError,
    UnexpectedProxyUsageError,
    _autocast,
    _joinNamePath,
    _typeStr,
)

KeyTypeVar = TypeVar("KeyTypeVar")
ItemTypeVar = TypeVar("ItemTypeVar")


class Dict(collections.abc.MutableMapping[KeyTypeVar, ItemTypeVar]):
    """An internal mapping container.

    This class emulates a `dict`, but adds validation and provenance.

    Parameters
    ----------
    config : `~lsst.pex.config.Config`
        Config to proxy.
    field : `~lsst.pex.config.DictField`
        Field to use.
    value : `~typing.Any`
        Value to store.
    at : `list` of `~lsst.pex.config.callStack.StackFrame`
        Stack frame for history recording. Will be calculated if `None`.
    label : `str`, optional
        Label to use for history recording.
    setHistory : `bool`, optional
        Whether to append to the history record.
    """

    def __init__(
        self,
        config: Config,
        field: DictField,
        value: Mapping[KeyTypeVar, ItemTypeVar],
        *,
        at: list[StackFrame] | None,
        label: str,
        setHistory: bool = True,
    ):
        self._field = field
        self._config_ = weakref.ref(config)
        self._dict = {}
        self._history = self._config._history.setdefault(self._field.name, [])
        self.__doc__ = field.doc
        if value is not None:
            try:
                for k in value:
                    # do not set history per-item
                    self.__setitem__(k, value[k], at=at, label=label, setHistory=False)
            except TypeError as e:
                msg = f"Value {value} is of incorrect type {_typeStr(value)}. Mapping type expected."
                raise FieldValidationError(self._field, self._config, msg) from e
        if setHistory:
            self._history.append((dict(self._dict), at, label))

    @property
    def _config(self) -> Config:
        # Config Fields should never outlive their config class instance
        # assert that as such here
        value = self._config_()
        assert value is not None
        return value

    history = property(lambda x: x._history)
    """History (read-only).
    """

    def _copy(self, config: Config) -> Dict:
        return type(self)(config, self._field, self._dict.copy(), at=None, label="copy", setHistory=False)

    def __getitem__(self, k: KeyTypeVar) -> ItemTypeVar:
        return self._dict[k]

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> Iterator[KeyTypeVar]:
        return iter(self._dict)

    def __contains__(self, k: Any) -> bool:
        return k in self._dict

    def __setitem__(
        self, k: KeyTypeVar, x: ItemTypeVar, at: Any = None, label: str = "setitem", setHistory: bool = True
    ) -> None:
        if self._config._frozen:
            msg = f"Cannot modify a frozen Config. Attempting to set item at key {k!r} to value {x}"
            raise FieldValidationError(self._field, self._config, msg)

        # validate keytype
        k = _autocast(k, self._field.keytype)
        if type(k) is not self._field.keytype:
            msg = f"Key {k!r} is of type {_typeStr(k)}, expected type {_typeStr(self._field.keytype)}"
            raise FieldValidationError(self._field, self._config, msg)

        # validate itemtype
        x = _autocast(x, self._field.itemtype)
        if self._field.itemtype is None:
            if type(x) not in self._field.supportedTypes and x is not None:
                msg = f"Value {x} at key {k!r} is of invalid type {_typeStr(x)}"
                raise FieldValidationError(self._field, self._config, msg)
        else:
            if type(x) is not self._field.itemtype and x is not None:
                msg = (
                    f"Value {x} at key {k!r} is of incorrect type {_typeStr(x)}. "
                    f"Expected type {_typeStr(self._field.itemtype)}"
                )
                raise FieldValidationError(self._field, self._config, msg)

        # validate key using keycheck
        if self._field.keyCheck is not None and not self._field.keyCheck(k):
            msg = f"Key {k!r} is not a valid key"
            raise FieldValidationError(self._field, self._config, msg)

        # validate item using itemcheck
        if self._field.itemCheck is not None and not self._field.itemCheck(x):
            msg = f"Item at key {k!r} is not a valid value: {x}"
            raise FieldValidationError(self._field, self._config, msg)

        if at is None:
            at = getCallStack()

        self._dict[k] = x
        if setHistory:
            self._history.append((dict(self._dict), at, label))

    def __delitem__(
        self, k: KeyTypeVar, at: Any = None, label: str = "delitem", setHistory: bool = True
    ) -> None:
        if self._config._frozen:
            raise FieldValidationError(self._field, self._config, "Cannot modify a frozen Config")

        del self._dict[k]
        if setHistory:
            if at is None:
                at = getCallStack()
            self._history.append((dict(self._dict), at, label))

    def __repr__(self):
        return repr(self._dict)

    def __str__(self):
        return str(self._dict)

    def __setattr__(self, attr, value, at=None, label="assignment"):
        if hasattr(getattr(self.__class__, attr, None), "__set__"):
            # This allows properties to work.
            object.__setattr__(self, attr, value)
        elif attr in self.__dict__ or attr in ["_field", "_config_", "_history", "_dict", "__doc__"]:
            # This allows specific private attributes to work.
            object.__setattr__(self, attr, value)
        else:
            # We throw everything else.
            msg = f"{_typeStr(self._field)} has no attribute {attr}"
            raise FieldValidationError(self._field, self._config, msg)

    def __reduce__(self):
        raise UnexpectedProxyUsageError(
            f"Proxy container for config field {self._field.name} cannot "
            "be pickled; it should be converted to a built-in container before "
            "being assigned to other objects or variables."
        )


class DictField(Field[Dict[KeyTypeVar, ItemTypeVar]], Generic[KeyTypeVar, ItemTypeVar]):
    """A configuration field (`~lsst.pex.config.Field` subclass) that maps keys
    and values.

    The types of both items and keys are restricted to these builtin types:
    `int`, `float`, `complex`, `bool`, and `str`). All keys share the same type
    and all values share the same type. Keys can have a different type from
    values.

    Parameters
    ----------
    doc : `str`
        A documentation string that describes the configuration field.
    keytype : {`int`, `float`, `complex`, `bool`, `str`}, optional
        The type of the mapping keys. All keys must have this type. Optional
        if keytype and itemtype are supplied as typing arguments to the class.
    itemtype : {`int`, `float`, `complex`, `bool`, `str`}, optional
        Type of the mapping values. Optional if keytype and itemtype are
        supplied as typing arguments to the class.
    default : `dict`, optional
        The default mapping.
    optional : `bool`, optional
        If `True`, the field doesn't need to have a set value.
    dictCheck : callable
        A function that validates the dictionary as a whole.
    keyCheck : callable
        A function that validates individual mapping keys.
    itemCheck : callable
        A function that validates individual mapping values.
    deprecated : None or `str`, optional
        A description of why this Field is deprecated, including removal date.
        If not None, the string is appended to the docstring for this Field.

    See Also
    --------
    ChoiceField
    ConfigChoiceField
    ConfigDictField
    ConfigField
    ConfigurableField
    Field
    ListField
    RangeField
    RegistryField

    Examples
    --------
    This field maps has `str` keys and `int` values:

    >>> from lsst.pex.config import Config, DictField
    >>> class MyConfig(Config):
    ...     field = DictField(
    ...         doc="Example string-to-int mapping field.",
    ...         keytype=str,
    ...         itemtype=int,
    ...         default={},
    ...     )
    >>> config = MyConfig()
    >>> config.field["myKey"] = 42
    >>> print(config.field)
    {'myKey': 42}
    """

    DictClass: type[Dict] = Dict

    @staticmethod
    def _parseTypingArgs(
        params: tuple[type, ...] | tuple[str, ...], kwds: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        if len(params) != 2:
            raise ValueError("Only tuples of types that are length 2 are supported")
        resultParams = []
        for typ in params:
            if isinstance(typ, str):
                _typ = ForwardRef(typ)
                # type ignore below because typeshed seems to be wrong. It
                # indicates there are only 2 args, as it was in python 3.8, but
                # 3.9+ takes 3 args.
                result = _typ._evaluate(globals(), locals(), recursive_guard=set())  # type: ignore
                if result is None:
                    raise ValueError("Could not deduce type from input")
                typ = cast(type, result)
            resultParams.append(typ)
        keyType, itemType = resultParams
        results = dict(kwds)
        if (supplied := kwds.get("keytype")) and supplied != keyType:
            raise ValueError("Conflicting definition for keytype")
        else:
            results["keytype"] = keyType
        if (supplied := kwds.get("itemtype")) and supplied != itemType:
            raise ValueError("Conflicting definition for itemtype")
        else:
            results["itemtype"] = itemType
        return results

    def __init__(
        self,
        doc,
        keytype=None,
        itemtype=None,
        default=None,
        optional=False,
        dictCheck=None,
        keyCheck=None,
        itemCheck=None,
        deprecated=None,
    ):
        source = getStackFrame()
        self._setup(
            doc=doc,
            dtype=Dict,
            default=default,
            check=None,
            optional=optional,
            source=source,
            deprecated=deprecated,
        )
        if keytype is None:
            raise ValueError(
                "keytype must either be supplied as an argument or as a type argument to the class"
            )
        if keytype not in self.supportedTypes:
            raise ValueError(f"'keytype' {_typeStr(keytype)} is not a supported type")
        elif itemtype is not None and itemtype not in self.supportedTypes:
            raise ValueError(f"'itemtype' {_typeStr(itemtype)} is not a supported type")

        check_errors = []
        for name, check in (("dictCheck", dictCheck), ("keyCheck", keyCheck), ("itemCheck", itemCheck)):
            if check is not None and not callable(check):
                check_errors.append(name)
        if check_errors:
            raise ValueError(f"{', '.join(check_errors)} must be callable")

        self.keytype = keytype
        self.itemtype = itemtype
        self.dictCheck = dictCheck
        self.keyCheck = keyCheck
        self.itemCheck = itemCheck

    def validate(self, instance):
        """Validate the field's value (for internal use only).

        Parameters
        ----------
        instance : `lsst.pex.config.Config`
            The configuration that contains this field.

        Raises
        ------
        lsst.pex.config.FieldValidationError
            Raised if validation fails for this field (see *Notes*).

        Notes
        -----
        This method validates values according to the following criteria:

        - A non-optional field is not `None`.
        - If a value is not `None`, it must pass the `ConfigField.dictCheck`
          user callback function.

        Individual key and item checks by the ``keyCheck`` and ``itemCheck``
        user callback functions are done immediately when the value is set on a
        key. Those checks are not repeated by this method.
        """
        Field.validate(self, instance)
        value = self.__get__(instance)
        if value is not None and self.dictCheck is not None and not self.dictCheck(value):
            msg = f"{value} is not a valid value"
            raise FieldValidationError(self, instance, msg)

    def __set__(
        self,
        instance: Config,
        value: Mapping[KeyTypeVar, ItemTypeVar] | None,
        at: Any = None,
        label: str = "assignment",
    ) -> None:
        if instance._frozen:
            msg = f"Cannot modify a frozen Config. Attempting to set field to value {value}"
            raise FieldValidationError(self, instance, msg)

        if at is None:
            at = getCallStack()
        if value is not None:
            value = self.DictClass(instance, self, value, at=at, label=label)
        else:
            history = instance._history.setdefault(self.name, [])
            history.append((value, at, label))

        instance._storage[self.name] = value

    def toDict(self, instance):
        """Convert this field's key-value pairs into a regular `dict`.

        Parameters
        ----------
        instance : `lsst.pex.config.Config`
            The configuration that contains this field.

        Returns
        -------
        result : `dict` or `None`
            If this field has a value of `None`, then this method returns
            `None`. Otherwise, this method returns the field's value as a
            regular Python `dict`.
        """
        value = self.__get__(instance)
        return dict(value) if value is not None else None

    def _copy_storage(self, old: Config, new: Config) -> Dict[KeyTypeVar, ItemTypeVar] | None:
        value: Dict[KeyTypeVar, ItemTypeVar] | None = old._storage[self.name]
        if value is not None:
            return value._copy(new)
        else:
            return None

    def _compare(self, instance1, instance2, shortcut, rtol, atol, output):
        """Compare two fields for equality.

        Used by `lsst.pex.ConfigDictField.compare`.

        Parameters
        ----------
        instance1 : `lsst.pex.config.Config`
            Left-hand side config instance to compare.
        instance2 : `lsst.pex.config.Config`
            Right-hand side config instance to compare.
        shortcut : `bool`
            If `True`, this function returns as soon as an inequality if found.
        rtol : `float`
            Relative tolerance for floating point comparisons.
        atol : `float`
            Absolute tolerance for floating point comparisons.
        output : callable
            A callable that takes a string, used (possibly repeatedly) to
            report inequalities.

        Returns
        -------
        isEqual : bool
            `True` if the fields are equal, `False` otherwise.

        Notes
        -----
        Floating point comparisons are performed by `numpy.allclose`.
        """
        d1 = getattr(instance1, self.name)
        d2 = getattr(instance2, self.name)
        name = getComparisonName(
            _joinNamePath(instance1._name, self.name), _joinNamePath(instance2._name, self.name)
        )
        if d1 is None or d2 is None:
            return compareScalars(name, d1, d2, output=output)
        if not compareScalars(f"{name} (keys)", set(d1.keys()), set(d2.keys()), output=output):
            return False
        equal = True
        for k, v1 in d1.items():
            v2 = d2[k]
            result = compareScalars(
                f"{name}[{k!r}]", v1, v2, dtype=self.itemtype, rtol=rtol, atol=atol, output=output
            )
            if not result and shortcut:
                return False
            equal = equal and result
        return equal



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("ConfigurableActionStruct", "ConfigurableActionStructField")

import weakref
from collections.abc import Iterable, Iterator, Mapping
from types import GenericAlias, SimpleNamespace
from typing import Any, Generic, TypeVar, overload

from lsst.pex.config.callStack import StackFrame, getCallStack, getStackFrame
from lsst.pex.config.comparison import compareConfigs, compareScalars, getComparisonName
from lsst.pex.config.config import Config, Field, FieldValidationError, _joinNamePath, _typeStr

from . import ActionTypeVar, ConfigurableAction


class ConfigurableActionStructUpdater:
    """Abstract the logic of using a dictionary to update a
    `ConfigurableActionStruct` through attribute assignment.

    This is useful in the context of setting configuration through pipelines
    or on the command line.
    """

    def __set__(
        self,
        instance: ConfigurableActionStruct,
        value: Mapping[str, ConfigurableAction] | ConfigurableActionStruct,
    ) -> None:
        if isinstance(value, Mapping):
            pass
        elif isinstance(value, ConfigurableActionStruct):
            # If the update target is a ConfigurableActionStruct, get the
            # internal dictionary
            value = value._attrs
        else:
            raise ValueError(
                "Can only update a ConfigurableActionStruct with an instance of such, or a mapping"
            )
        for name, action in value.items():
            setattr(instance, name, action)

    def __get__(self, instance, objtype=None) -> None:
        # This descriptor does not support fetching any value
        return None


class ConfigurableActionStructRemover:
    """Abstract the logic of removing an iterable of action names from a
    `ConfigurableActionStruct` at one time using attribute assignment.

    This is useful in the context of setting configuration through pipelines
    or on the command line.

    Raises
    ------
    AttributeError
        Raised if an attribute specified for removal does not exist in the
        ConfigurableActionStruct
    """

    def __set__(self, instance: ConfigurableActionStruct, value: str | Iterable[str]) -> None:
        # strings are iterable, but not in the way that is intended. If a
        # single name is specified, turn it into a tuple before attempting
        # to remove the attribute
        if isinstance(value, str):
            value = (value,)
        for name in value:
            delattr(instance, name)

    def __get__(self, instance, objtype=None) -> None:
        # This descriptor does not support fetching any value
        return None


class ConfigurableActionStruct(Generic[ActionTypeVar]):
    """A ConfigurableActionStruct is the storage backend class that supports
    the ConfigurableActionStructField. This class should not be created
    directly.

    This class allows managing a collection of `ConfigurableAction` with a
    struct like interface, that is to say in an attribute like notation.

    Parameters
    ----------
    config : `~lsst.pex.config.Config`
        Config to use.
    field : `ConfigurableActionStructField`
        Field to use.
    value : `~collections.abc.Mapping` [`str`, `ConfigurableAction`]
        Value to assign.
    at : `list` of `~lsst.pex.config.callStack.StackFrame` or `None`, optional
        Stack frames to use for history recording.
    label : `str`, optional
        Label to use for history recording.

    Notes
    -----
    Attributes can be dynamically added or removed as such:

    .. code-block:: python

        ConfigurableActionStructInstance.variable1 = a_configurable_action
        del ConfigurableActionStructInstance.variable1

    Each action is then available to be individually configured as a normal
    `lsst.pex.config.Config` object.

    `ConfigurableActionStruct` supports two special convenience attributes.

    The first is ``update``. You may assign a dict of `ConfigurableAction` or a
    `ConfigurableActionStruct` to this attribute which will update the
    `ConfigurableActionStruct` on which the attribute is invoked such that it
    will be updated to contain the entries specified by the structure on the
    right hand side of the equals sign.

    The second convenience attribute is named ``remove``. You may assign an
    iterable of strings which correspond to attribute names on the
    `ConfigurableActionStruct`. All of the corresponding attributes will then
    be removed. If any attribute does not exist, an `AttributeError` will be
    raised. Any attributes in the Iterable prior to the name which raises will
    have been removed from the `ConfigurableActionStruct`
    """

    # declare attributes that are set with __setattr__
    _config_: weakref.ref
    _attrs: dict[str, ActionTypeVar]
    _field: ConfigurableActionStructField
    _history: list[tuple]

    # create descriptors to handle special update and remove behavior
    update = ConfigurableActionStructUpdater()
    remove = ConfigurableActionStructRemover()

    def __init__(
        self,
        config: Config,
        field: ConfigurableActionStructField,
        value: Mapping[str, ConfigurableAction],
        at: Any,
        label: str,
    ):
        object.__setattr__(self, "_config_", weakref.ref(config))
        object.__setattr__(self, "_attrs", {})
        object.__setattr__(self, "_field", field)
        object.__setattr__(self, "_history", [])

        if at is not None:
            self.history.append(("Struct initialized", at, label))

        if value is not None:
            for k, v in value.items():
                setattr(self, k, v)

    def _copy(self, config: Config) -> ConfigurableActionStruct:
        result = ConfigurableActionStruct(config, self._field, self._attrs, at=None, label="copy")
        result.history.extend(self.history)
        return result

    @property
    def _config(self) -> Config:
        # Config Fields should never outlive their config class instance
        # assert that as such here
        value = self._config_()
        assert value is not None
        return value

    @property
    def history(self) -> list[tuple]:
        return self._history

    @property
    def fieldNames(self) -> Iterable[str]:
        return self._attrs.keys()

    def __setattr__(
        self,
        attr: str,
        value: ActionTypeVar | type[ActionTypeVar],
        at=None,
        label="setattr",
        setHistory=False,
    ) -> None:
        if hasattr(self._config, "_frozen") and self._config._frozen:
            msg = f"Cannot modify a frozen Config. Attempting to set item {attr} to value {value}"
            raise FieldValidationError(self._field, self._config, msg)

        # verify that someone has not passed a string with a space or leading
        # number or something through the dict assignment update interface
        if not attr.isidentifier():
            raise ValueError("Names used in ConfigurableStructs must be valid as python variable names")

        if attr not in (self.__dict__.keys() | type(self).__dict__.keys()):
            base_name = _joinNamePath(self._config._name, self._field.name)
            name = _joinNamePath(base_name, attr)
            if at is None:
                at = getCallStack()
            if isinstance(value, ConfigurableAction):
                valueInst = type(value)(__name=name, __at=at, __label=label, **value._storage)
            else:
                valueInst = value(__name=name, __at=at, __label=label)
            self._attrs[attr] = valueInst
        else:
            super().__setattr__(attr, value)

    def __getattr__(self, attr) -> Any:
        if attr in object.__getattribute__(self, "_attrs"):
            result = self._attrs[attr]
            result.identity = attr
            return result
        else:
            super().__getattribute__(attr)

    def __delattr__(self, name):
        if name in self._attrs:
            del self._attrs[name]
        else:
            super().__delattr__(name)

    def __iter__(self) -> Iterator[ActionTypeVar]:
        for name in self.fieldNames:
            yield getattr(self, name)

    def items(self) -> Iterable[tuple[str, ActionTypeVar]]:
        for name in self.fieldNames:
            yield name, getattr(self, name)

    def __bool__(self) -> bool:
        return bool(self._attrs)


T = TypeVar("T", bound="ConfigurableActionStructField")


class ConfigurableActionStructField(Field[ActionTypeVar]):
    """`ConfigurableActionStructField` is a `~lsst.pex.config.Field` subclass
    that allows a `ConfigurableAction` to be organized in a
    `~lsst.pex.config.Config` class in a manner similar to how a
    `~lsst.pipe.base.Struct` works.

    This class uses a `ConfigurableActionStruct` as an intermediary object to
    organize the `ConfigurableAction`. See its documentation for further
    information.

    Parameters
    ----------
    doc : `str`
        Documentation string.
    default : `~collections.abc.Mapping` [ `str`, `ConfigurableAction` ] \
            or `None`, optional
        Default value.
    optional : `bool`, optional
        If `True`, the field doesn't need to have a set value.
    deprecated : `bool` or `None`, optional
        A description of why this Field is deprecated, including removal date.
        If not `None`, the string is appended to the docstring for this Field.
    """

    # specify StructClass to make this more generic for potential future
    # inheritance
    StructClass = ConfigurableActionStruct

    # Explicitly annotate these on the class, they are present in the base
    # class through injection, so type systems have trouble seeing them.
    name: str
    default: Mapping[str, ConfigurableAction] | None

    def __init__(
        self,
        doc: str,
        default: Mapping[str, ConfigurableAction] | None = None,
        optional: bool = False,
        deprecated=None,
    ):
        source = getStackFrame()
        self._setup(
            doc=doc,
            dtype=self.__class__,
            default=default,
            check=None,
            optional=optional,
            source=source,
            deprecated=deprecated,
        )

    def __class_getitem__(cls, params):
        return GenericAlias(cls, params)

    def __set__(
        self,
        instance: Config,
        value: (
            None
            | Mapping[str, ConfigurableAction]
            | SimpleNamespace
            | ConfigurableActionStruct
            | ConfigurableActionStructField
            | type[ConfigurableActionStructField]
        ),
        at: Iterable[StackFrame] = None,
        label: str = "assigment",
    ):
        if instance._frozen:
            msg = f"Cannot modify a frozen Config. Attempting to set field to value {value}"
            raise FieldValidationError(self, instance, msg)

        if at is None:
            at = getCallStack()

        if value is None or (self.default is not None and self.default == value):
            value = self.StructClass(instance, self, value, at=at, label=label)
        else:
            # An actual value is being assigned check for what it is
            if isinstance(value, self.StructClass):
                # If this is a ConfigurableActionStruct, we need to make our
                # own copy that references this current field
                value = self.StructClass(instance, self, value._attrs, at=at, label=label)
            elif isinstance(value, SimpleNamespace):
                # If this is a a python analogous container, we need to make
                # a ConfigurableActionStruct initialized with this data
                value = self.StructClass(instance, self, vars(value), at=at, label=label)

            elif type(value) is ConfigurableActionStructField:
                raise ValueError(
                    "ConfigurableActionStructFields can only be used in a class body declaration"
                    f"Use a {self.StructClass}, SimpleNamespace or Struct"
                )
            else:
                raise ValueError(f"Unrecognized value {value}, cannot be assigned to this field")

            history = instance._history.setdefault(self.name, [])
            history.append((value, at, label))

        if not isinstance(value, ConfigurableActionStruct):
            raise FieldValidationError(
                self, instance, "Can only assign things that are subclasses of Configurable Action"
            )
        instance._storage[self.name] = value

    @overload
    def __get__(
        self, instance: None, owner: Any = None, at: Any = None, label: str = "default"
    ) -> ConfigurableActionStruct[ActionTypeVar]: ...

    @overload
    def __get__(
        self, instance: Config, owner: Any = None, at: Any = None, label: str = "default"
    ) -> ConfigurableActionStruct[ActionTypeVar]: ...

    def __get__(self, instance, owner=None, at=None, label="default"):
        if instance is None or not isinstance(instance, Config):
            return self
        else:
            field: ConfigurableActionStruct | None = instance._storage[self.name]
            return field

    def rename(self, instance: Config):
        actionStruct: ConfigurableActionStruct = self.__get__(instance)
        if actionStruct is not None:
            for k, v in actionStruct.items():
                base_name = _joinNamePath(instance._name, self.name)
                fullname = _joinNamePath(base_name, k)
                v._rename(fullname)

    def validate(self, instance: Config):
        value = self.__get__(instance)
        if value is not None:
            for item in value:
                item.validate()

    def toDict(self, instance):
        actionStruct = self.__get__(instance)
        if actionStruct is None:
            return None

        dict_ = {k: v.toDict() for k, v in actionStruct.items()}

        return dict_

    def _copy_storage(self, old: Config, new: Config) -> ConfigurableActionStruct:
        struct: ConfigurableActionStruct | None = old._storage.get(self.name)
        if struct is not None:
            return struct._copy(new)
        else:
            return None

    def save(self, outfile, instance):
        actionStruct = self.__get__(instance)
        fullname = _joinNamePath(instance._name, self.name)

        # Ensure that a struct is always empty before assigning to it.
        outfile.write(f"{fullname}=None\n")

        if actionStruct is None:
            return

        for _, v in sorted(actionStruct.items()):
            outfile.write(f"{v._name}={_typeStr(v)}()\n")
            v._save(outfile)

    def freeze(self, instance):
        actionStruct = self.__get__(instance)
        if actionStruct is not None:
            for v in actionStruct:
                v.freeze()

    def _collectImports(self, instance, imports):
        # docstring inherited from Field
        actionStruct = self.__get__(instance)
        for v in actionStruct:
            v._collectImports()
            imports |= v._imports

    def _compare(self, instance1, instance2, shortcut, rtol, atol, output):
        """Compare two fields for equality.

        Parameters
        ----------
        instance1 : `lsst.pex.config.Config`
            Left-hand side config instance to compare.
        instance2 : `lsst.pex.config.Config`
            Right-hand side config instance to compare.
        shortcut : `bool`
            If `True`, this function returns as soon as an inequality if found.
        rtol : `float`
            Relative tolerance for floating point comparisons.
        atol : `float`
            Absolute tolerance for floating point comparisons.
        output : callable
            A callable that takes a string, used (possibly repeatedly) to
            report inequalities.

        Returns
        -------
        isEqual : bool
            `True` if the fields are equal, `False` otherwise.

        Notes
        -----
        Floating point comparisons are performed by `numpy.allclose`.
        """
        d1: ConfigurableActionStruct = getattr(instance1, self.name)
        d2: ConfigurableActionStruct = getattr(instance2, self.name)
        name = getComparisonName(
            _joinNamePath(instance1._name, self.name), _joinNamePath(instance2._name, self.name)
        )
        if not compareScalars(f"{name} (fields)", set(d1.fieldNames), set(d2.fieldNames), output=output):
            return False
        equal = True
        for k, v1 in d1.items():
            v2 = getattr(d2, k)
            result = compareConfigs(
                f"{name}.{k}", v1, v2, shortcut=shortcut, rtol=rtol, atol=atol, output=output
            )
            if not result and shortcut:
                return False
            equal = equal and result
        return equal



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

import lsst.pex.config as pexConf


class Config1(pexConf.Config):
    """First test config."""

    f = pexConf.Field("Config1.f", float, default=4)


class Config2(Config1):
    """Second test config."""

    def setDefaults(self):
        self.f = 5


class Config3(Config1):
    """Third test config."""

    def __init__(self, **kw):
        self.f = 6


class SquashingDefaultsTest(unittest.TestCase):
    """Tests for squashing defaults."""

    def test(self):
        c1 = Config1()
        self.assertEqual(c1.f, 4)
        c1 = Config1(f=9)
        self.assertEqual(c1.f, 9)

        c2 = Config2()
        self.assertEqual(c2.f, 5)
        c2 = Config2(f=10)
        self.assertEqual(c2.f, 10)

        c3 = Config3()
        self.assertEqual(c3.f, 6)
        c3 = Config3(f=11)
        self.assertEqual(c3.f, 6)


if __name__ == "__main__":
    unittest.main()
